@@ -1,13 +1,18 @@
 from mongoengine.errors import NotRegistered
 
-__all__ = ('ALLOW_INHERITANCE', 'get_document', '_document_registry')
+__all__ = ('UPDATE_OPERATORS', 'get_document', '_document_registry')
+
+
+UPDATE_OPERATORS = set(['set', 'unset', 'inc', 'dec', 'pop', 'push',
+                        'push_all', 'pull', 'pull_all', 'add_to_set',
+                        'set_on_insert', 'min', 'max'])
 
-ALLOW_INHERITANCE = False
 
 _document_registry = {}
 
 
 def get_document(name):
+    """Get a document class by name."""
     doc = _document_registry.get(name, None)
     if not doc:
         # Possible old style name
