@@ -4,22 +4,17 @@ import weakref
 
 from bson import DBRef, ObjectId, SON
 import pymongo
+import six
 
+from mongoengine.base.common import UPDATE_OPERATORS
+from mongoengine.base.datastructures import (BaseDict, BaseList,
+                                             EmbeddedDocumentList)
 from mongoengine.common import _import_class
 from mongoengine.errors import ValidationError
-from mongoengine.base.common import ALLOW_INHERITANCE
-from mongoengine.base.datastructures import (
-    BaseDict, BaseList, EmbeddedDocumentList
-)
-import collections
-
-__all__ = ("BaseField", "ComplexBaseField",
-           "ObjectIdField", "GeoJsonBaseField")
 
 
-UPDATE_OPERATORS = set(['set', 'unset', 'inc', 'dec', 'pop', 'push',
-                        'push_all', 'pull', 'pull_all', 'add_to_set',
-                        'set_on_insert', 'min', 'max'])
+__all__ = ('BaseField', 'ComplexBaseField', 'ObjectIdField',
+           'GeoJsonBaseField')
 
 
 class BaseField(object):
@@ -28,7 +23,6 @@ class BaseField(object):
 
     .. versionchanged:: 0.5 - added verbose and help text
     """
-
     name = None
     _geo_index = False
     _auto_gen = False  # Call `generate` to generate a value
@@ -74,7 +68,7 @@ class BaseField(object):
         self.db_field = (db_field or name) if not primary_key else '_id'
 
         if name:
-            msg = "Fields' 'name' attribute deprecated in favour of 'db_field'"
+            msg = 'Field\'s "name" attribute deprecated in favour of "db_field"'
             warnings.warn(msg, DeprecationWarning)
         self.required = required or primary_key
         self.default = default
@@ -86,13 +80,13 @@ class BaseField(object):
         self.null = null
         self.sparse = sparse
         self._owner_document = None
-        
+
         # Detect and report conflicts between metadata and base properties.
         conflicts = set(dir(self)) & set(kwargs)
         if conflicts:
-            raise TypeError("%s already has attribute(s): %s" % (
-                self.__class__.__name__, ', '.join(conflicts) ))
-        
+            raise TypeError('%s already has attribute(s): %s' % (
+                self.__class__.__name__, ', '.join(conflicts)))
+
         # Assign metadata to the instance
         # This efficient method is available because no __slots__ are defined.
         self.__dict__.update(kwargs)
@@ -126,7 +120,7 @@ class BaseField(object):
                 value = None
             elif self.default is not None:
                 value = self.default
-                if isinstance(value, collections.Callable):
+                if callable(value):
                     value = value()
 
         if instance._initialised:
@@ -134,7 +128,7 @@ class BaseField(object):
                 if (self.name not in instance._data or
                         instance._data[self.name] != value):
                     instance._mark_as_changed(self.name)
-            except:
+            except Exception:
                 # Values cant be compared eg: naive and tz datetimes
                 # So mark it as changed
                 instance._mark_as_changed(self.name)
@@ -148,32 +142,39 @@ class BaseField(object):
                     v._instance = weakref.proxy(instance)
         instance._data[self.name] = value
 
-    def error(self, message="", errors=None, field_name=None):
-        """Raises a ValidationError.
-        """
+    def error(self, message='', errors=None, field_name=None):
+        """Raise a ValidationError."""
         field_name = field_name if field_name else self.name
         raise ValidationError(message, errors=errors, field_name=field_name)
 
     def to_python(self, value):
-        """Convert a MongoDB-compatible type to a Python type.
-        """
+        """Convert a MongoDB-compatible type to a Python type."""
         return value
 
     def to_mongo(self, value):
-        """Convert a Python type to a MongoDB-compatible type.
-        """
+        """Convert a Python type to a MongoDB-compatible type."""
         return self.to_python(value)
 
+    def _to_mongo_safe_call(self, value, use_db_field=True, fields=None):
+        """Helper method to call to_mongo with proper inputs."""
+        f_inputs = self.to_mongo.__code__.co_varnames
+        ex_vars = {}
+        if 'fields' in f_inputs:
+            ex_vars['fields'] = fields
+
+        if 'use_db_field' in f_inputs:
+            ex_vars['use_db_field'] = use_db_field
+
+        return self.to_mongo(value, **ex_vars)
+
     def prepare_query_value(self, op, value):
-        """Prepare a value that is being used in a query for PyMongo.
-        """
+        """Prepare a value that is being used in a query for PyMongo."""
         if op in UPDATE_OPERATORS:
             self.validate(value)
         return value
 
     def validate(self, value, clean=True):
-        """Perform validation on a value.
-        """
+        """Perform validation on a value."""
         pass
 
     def _validate_choices(self, value):
@@ -188,12 +189,13 @@ class BaseField(object):
         if isinstance(value, (Document, EmbeddedDocument)):
             if not any(isinstance(value, c) for c in choice_list):
                 self.error(
-                    'Value must be instance of %s' % str(choice_list)
+                    'Value must be an instance of %s' % (
+                        six.text_type(choice_list)
+                    )
                 )
         # Choices which are types other than Documents
         elif value not in choice_list:
-            self.error('Value must be one of %s' % str(choice_list))
-
+            self.error('Value must be one of %s' % six.text_type(choice_list))
 
     def _validate(self, value, **kwargs):
         # Check the Choices Constraint
@@ -202,7 +204,7 @@ class BaseField(object):
 
         # check validation argument
         if self.validation is not None:
-            if isinstance(self.validation, collections.Callable):
+            if callable(self.validation):
                 if not self.validation(value):
                     self.error('Value does not match custom validation method')
             else:
@@ -236,8 +238,7 @@ class ComplexBaseField(BaseField):
     field = None
 
     def __get__(self, instance, owner):
-        """Descriptor to automatically dereference references.
-        """
+        """Descriptor to automatically dereference references."""
         if instance is None:
             # Document class being used rather than a document object
             return self
@@ -249,7 +250,7 @@ class ComplexBaseField(BaseField):
                        (self.field is None or isinstance(self.field,
                                                          (GenericReferenceField, ReferenceField))))
 
-        _dereference = _import_class("DeReference")()
+        _dereference = _import_class('DeReference')()
 
         self._auto_dereference = instance._fields[self.name]._auto_dereference
         if instance._initialised and dereference and instance._data.get(self.name):
@@ -284,11 +285,8 @@ class ComplexBaseField(BaseField):
         return value
 
     def to_python(self, value):
-        """Convert a MongoDB-compatible type to a Python type.
-        """
-        Document = _import_class('Document')
-
-        if isinstance(value, str):
+        """Convert a MongoDB-compatible type to a Python type."""
+        if isinstance(value, six.string_types):
             return value
 
         if hasattr(value, 'to_python'):
@@ -298,15 +296,16 @@ class ComplexBaseField(BaseField):
         if not hasattr(value, 'items'):
             try:
                 is_list = True
-                value = dict([(k, v) for k, v in enumerate(value)])
+                value = {k: v for k, v in enumerate(value)}
             except TypeError:  # Not iterable return the value
                 return value
 
         if self.field:
             self.field._auto_dereference = self._auto_dereference
-            value_dict = dict([(key, self.field.to_python(item))
-                               for key, item in list(value.items())])
+            value_dict = {key: self.field.to_python(item)
+                          for key, item in list(value.items())}
         else:
+            Document = _import_class('Document')
             value_dict = {}
             for k, v in list(value.items()):
                 if isinstance(v, Document):
@@ -326,21 +325,20 @@ class ComplexBaseField(BaseField):
                                          key=operator.itemgetter(0))]
         return value_dict
 
-    def to_mongo(self, value):
-        """Convert a Python type to a MongoDB-compatible type.
-        """
-        Document = _import_class("Document")
-        EmbeddedDocument = _import_class("EmbeddedDocument")
-        GenericReferenceField = _import_class("GenericReferenceField")
+    def to_mongo(self, value, use_db_field=True, fields=None):
+        """Convert a Python type to a MongoDB-compatible type."""
+        Document = _import_class('Document')
+        EmbeddedDocument = _import_class('EmbeddedDocument')
+        GenericReferenceField = _import_class('GenericReferenceField')
 
-        if isinstance(value, str):
+        if isinstance(value, six.string_types):
             return value
 
         if hasattr(value, 'to_mongo'):
             if isinstance(value, Document):
                 return GenericReferenceField().to_mongo(value)
             cls = value.__class__
-            val = value.to_mongo()
+            val = value.to_mongo(use_db_field, fields)
             # If it's a document that is not inherited add _cls
             if isinstance(value, EmbeddedDocument):
                 val['_cls'] = cls.__name__
@@ -350,13 +348,15 @@ class ComplexBaseField(BaseField):
         if not hasattr(value, 'items'):
             try:
                 is_list = True
-                value = dict([(k, v) for k, v in enumerate(value)])
+                value = {k: v for k, v in enumerate(value)}
             except TypeError:  # Not iterable return the value
                 return value
 
         if self.field:
-            value_dict = dict([(key, self.field.to_mongo(item))
-                               for key, item in value.items()])
+            value_dict = {
+                key: self.field._to_mongo_safe_call(item, use_db_field, fields)
+                for key, item in value.items()
+            }
         else:
             value_dict = {}
             for k, v in value.items():
@@ -370,9 +370,7 @@ class ComplexBaseField(BaseField):
                     # any _cls data so make it a generic reference allows
                     # us to dereference
                     meta = getattr(v, '_meta', {})
-                    allow_inheritance = (
-                        meta.get('allow_inheritance', ALLOW_INHERITANCE)
-                        is True)
+                    allow_inheritance = meta.get('allow_inheritance')
                     if not allow_inheritance and not self.field:
                         value_dict[k] = GenericReferenceField().to_mongo(v)
                     else:
@@ -380,13 +378,13 @@ class ComplexBaseField(BaseField):
                         value_dict[k] = DBRef(collection, v.pk)
                 elif hasattr(v, 'to_mongo'):
                     cls = v.__class__
-                    val = v.to_mongo()
+                    val = v.to_mongo(use_db_field, fields)
                     # If it's a document that is not inherited add _cls
                     if isinstance(v, (Document, EmbeddedDocument)):
                         val['_cls'] = cls.__name__
                     value_dict[k] = val
                 else:
-                    value_dict[k] = self.to_mongo(v)
+                    value_dict[k] = self.to_mongo(v, use_db_field, fields)
 
         if is_list:  # Convert back to a list
             return [v for _, v in sorted(list(value_dict.items()),
@@ -394,8 +392,7 @@ class ComplexBaseField(BaseField):
         return value_dict
 
     def validate(self, value):
-        """If field is provided ensure the value is valid.
-        """
+        """If field is provided ensure the value is valid."""
         errors = {}
         if self.field:
             if hasattr(value, 'iteritems') or hasattr(value, 'items'):
@@ -433,24 +430,23 @@ class ComplexBaseField(BaseField):
 
 
 class ObjectIdField(BaseField):
-    """A field wrapper around MongoDB's ObjectIds.
-    """
+    """A field wrapper around MongoDB's ObjectIds."""
 
     def to_python(self, value):
         try:
             if not isinstance(value, ObjectId):
                 value = ObjectId(value)
-        except:
+        except Exception:
             pass
         return value
 
     def to_mongo(self, value):
         if not isinstance(value, ObjectId):
             try:
-                return ObjectId(str(value))
+                return ObjectId(six.text_type(value))
             except Exception as e:
                 # e.message attribute has been deprecated since Python 2.6
-                self.error(str(e))
+                self.error(six.text_type(e))
         return value
 
     def prepare_query_value(self, op, value):
@@ -458,8 +454,8 @@ class ObjectIdField(BaseField):
 
     def validate(self, value):
         try:
-            ObjectId(str(value))
-        except:
+            ObjectId(six.text_type(value))
+        except Exception:
             self.error('Invalid Object ID')
 
 
@@ -470,21 +466,20 @@ class GeoJsonBaseField(BaseField):
     """
 
     _geo_index = pymongo.GEOSPHERE
-    _type = "GeoBase"
+    _type = 'GeoBase'
 
     def __init__(self, auto_index=True, *args, **kwargs):
         """
-        :param bool auto_index: Automatically create a "2dsphere" index.\
+        :param bool auto_index: Automatically create a '2dsphere' index.\
             Defaults to `True`.
         """
-        self._name = "%sField" % self._type
+        self._name = '%sField' % self._type
         if not auto_index:
             self._geo_index = False
         super(GeoJsonBaseField, self).__init__(*args, **kwargs)
 
     def validate(self, value):
-        """Validate the GeoJson object based on its type
-        """
+        """Validate the GeoJson object based on its type."""
         if isinstance(value, dict):
             if set(value.keys()) == set(['type', 'coordinates']):
                 if value['type'] != self._type:
@@ -499,7 +494,7 @@ class GeoJsonBaseField(BaseField):
             self.error('%s can only accept lists of [x, y]' % self._name)
             return
 
-        validate = getattr(self, "_validate_%s" % self._type.lower())
+        validate = getattr(self, '_validate_%s' % self._type.lower())
         error = validate(value)
         if error:
             self.error(error)
@@ -511,8 +506,8 @@ class GeoJsonBaseField(BaseField):
         # Quick and dirty validator
         try:
             value[0][0][0]
-        except:
-            return "Invalid Polygon must contain at least one valid linestring"
+        except (TypeError, IndexError):
+            return 'Invalid Polygon must contain at least one valid linestring'
 
         errors = []
         for val in value:
@@ -523,20 +518,20 @@ class GeoJsonBaseField(BaseField):
                 errors.append(error)
         if errors:
             if top_level:
-                return "Invalid Polygon:\n%s" % ", ".join(errors)
+                return 'Invalid Polygon:\n%s' % ', '.join(errors)
             else:
-                return "%s" % ", ".join(errors)
+                return '%s' % ', '.join(errors)
 
     def _validate_linestring(self, value, top_level=True):
-        """Validates a linestring"""
+        """Validate a linestring."""
         if not isinstance(value, (list, tuple)):
             return 'LineStrings must contain list of coordinate pairs'
 
         # Quick and dirty validator
         try:
             value[0][0]
-        except:
-            return "Invalid LineString must contain at least one valid point"
+        except (TypeError, IndexError):
+            return 'Invalid LineString must contain at least one valid point'
 
         errors = []
         for val in value:
@@ -545,19 +540,19 @@ class GeoJsonBaseField(BaseField):
                 errors.append(error)
         if errors:
             if top_level:
-                return "Invalid LineString:\n%s" % ", ".join(errors)
+                return 'Invalid LineString:\n%s' % ', '.join(errors)
             else:
-                return "%s" % ", ".join(errors)
+                return '%s' % ', '.join(errors)
 
     def _validate_point(self, value):
         """Validate each set of coords"""
         if not isinstance(value, (list, tuple)):
             return 'Points must be a list of coordinate pairs'
         elif not len(value) == 2:
-            return "Value (%s) must be a two-dimensional point" % repr(value)
+            return 'Value (%s) must be a two-dimensional point' % repr(value)
         elif (not isinstance(value[0], (float, int)) or
               not isinstance(value[1], (float, int))):
-            return "Both values (%s) in point must be float or int" % repr(value)
+            return 'Both values (%s) in point must be float or int' % repr(value)
 
     def _validate_multipoint(self, value):
         if not isinstance(value, (list, tuple)):
@@ -566,8 +561,8 @@ class GeoJsonBaseField(BaseField):
         # Quick and dirty validator
         try:
             value[0][0]
-        except:
-            return "Invalid MultiPoint must contain at least one valid point"
+        except (TypeError, IndexError):
+            return 'Invalid MultiPoint must contain at least one valid point'
 
         errors = []
         for point in value:
@@ -576,7 +571,7 @@ class GeoJsonBaseField(BaseField):
                 errors.append(error)
 
         if errors:
-            return "%s" % ", ".join(errors)
+            return '%s' % ', '.join(errors)
 
     def _validate_multilinestring(self, value, top_level=True):
         if not isinstance(value, (list, tuple)):
@@ -585,8 +580,8 @@ class GeoJsonBaseField(BaseField):
         # Quick and dirty validator
         try:
             value[0][0][0]
-        except:
-            return "Invalid MultiLineString must contain at least one valid linestring"
+        except (TypeError, IndexError):
+            return 'Invalid MultiLineString must contain at least one valid linestring'
 
         errors = []
         for linestring in value:
@@ -596,9 +591,9 @@ class GeoJsonBaseField(BaseField):
 
         if errors:
             if top_level:
-                return "Invalid MultiLineString:\n%s" % ", ".join(errors)
+                return 'Invalid MultiLineString:\n%s' % ', '.join(errors)
             else:
-                return "%s" % ", ".join(errors)
+                return '%s' % ', '.join(errors)
 
     def _validate_multipolygon(self, value):
         if not isinstance(value, (list, tuple)):
@@ -607,8 +602,8 @@ class GeoJsonBaseField(BaseField):
         # Quick and dirty validator
         try:
             value[0][0][0][0]
-        except:
-            return "Invalid MultiPolygon must contain at least one valid Polygon"
+        except (TypeError, IndexError):
+            return 'Invalid MultiPolygon must contain at least one valid Polygon'
 
         errors = []
         for polygon in value:
@@ -617,9 +612,9 @@ class GeoJsonBaseField(BaseField):
                 errors.append(error)
 
         if errors:
-            return "Invalid MultiPolygon:\n%s" % ", ".join(errors)
+            return 'Invalid MultiPolygon:\n%s' % ', '.join(errors)
 
     def to_mongo(self, value):
         if isinstance(value, dict):
             return value
-        return SON([("type", self._type), ("coordinates", value)])
+        return SON([('type', self._type), ('coordinates', value)])
