@@ -1,29 +1,23 @@
-import warnings
-import pymongo
 import re
+import warnings
 
-from pymongo.read_preferences import ReadPreference
 from bson.dbref import DBRef
+import pymongo
+from pymongo.read_preferences import ReadPreference
+import six
+
 from mongoengine import signals
+from mongoengine.base import (BaseDict, BaseDocument, BaseList,
+                              DocumentMetaclass, EmbeddedDocumentList,
+                              TopLevelDocumentMetaclass, get_document)
 from mongoengine.common import _import_class
-from mongoengine.base import (
-    DocumentMetaclass,
-    TopLevelDocumentMetaclass,
-    BaseDocument,
-    BaseDict,
-    BaseList,
-    EmbeddedDocumentList,
-    ALLOW_INHERITANCE,
-    get_document
-)
-from mongoengine.errors import (InvalidQueryError, InvalidDocumentError,
+from mongoengine.connection import DEFAULT_CONNECTION_NAME, get_db
+from mongoengine.context_managers import switch_collection, switch_db
+from mongoengine.errors import (InvalidDocumentError, InvalidQueryError,
                                 SaveConditionError)
 from mongoengine.python_support import IS_PYMONGO_3
-from mongoengine.queryset import (OperationError, NotUniqueError,
+from mongoengine.queryset import (NotUniqueError, OperationError,
                                   QuerySet, transform)
-from mongoengine.connection import get_db, DEFAULT_CONNECTION_NAME
-from mongoengine.context_managers import switch_db, switch_collection
-import collections
 
 __all__ = ('Document', 'EmbeddedDocument', 'DynamicDocument',
            'DynamicEmbeddedDocument', 'OperationError',
@@ -31,12 +25,10 @@ __all__ = ('Document', 'EmbeddedDocument', 'DynamicDocument',
 
 
 def includes_cls(fields):
-    """ Helper function used for ensuring and comparing indexes
-    """
-
+    """Helper function used for ensuring and comparing indexes."""
     first_field = None
     if len(fields):
-        if isinstance(fields[0], str):
+        if isinstance(fields[0], six.string_types):
             first_field = fields[0]
         elif isinstance(fields[0], (list, tuple)) and len(fields[0]):
             first_field = fields[0][0]
@@ -57,9 +49,8 @@ class EmbeddedDocument(BaseDocument, metaclass=DocumentMetaclass):
     to create a specialised version of the embedded document that will be
     stored in the same collection. To facilitate this behaviour a `_cls`
     field is added to documents (hidden though the MongoEngine interface).
-    To disable this behaviour and remove the dependence on the presence of
-    `_cls` set :attr:`allow_inheritance` to ``False`` in the :attr:`meta`
-    dictionary.
+    To enable this behaviour set :attr:`allow_inheritance` to ``True`` in the
+    :attr:`meta` dictionary.
     """
 
     __slots__ = ('_instance', )
@@ -81,6 +72,15 @@ class EmbeddedDocument(BaseDocument, metaclass=DocumentMetaclass):
     def __ne__(self, other):
         return not self.__eq__(other)
 
+    def to_mongo(self, *args, **kwargs):
+        data = super(EmbeddedDocument, self).to_mongo(*args, **kwargs)
+
+        # remove _id from the SON if it's in it and it's None
+        if '_id' in data and data['_id'] is None:
+            del data['_id']
+
+        return data
+
     def save(self, *args, **kwargs):
         self._instance.save(*args, **kwargs)
 
@@ -105,9 +105,8 @@ class Document(BaseDocument, metaclass=TopLevelDocumentMetaclass):
     create a specialised version of the document that will be stored in the
     same collection. To facilitate this behaviour a `_cls`
     field is added to documents (hidden though the MongoEngine interface).
-    To disable this behaviour and remove the dependence on the presence of
-    `_cls` set :attr:`allow_inheritance` to ``False`` in the :attr:`meta`
-    dictionary.
+    To enable this behaviourset :attr:`allow_inheritance` to ``True`` in the
+    :attr:`meta` dictionary.
 
     A :class:`~mongoengine.Document` may use a **Capped Collection** by
     specifying :attr:`max_documents` and :attr:`max_size` in the :attr:`meta`
@@ -147,26 +146,22 @@ class Document(BaseDocument, metaclass=TopLevelDocumentMetaclass):
 
     __slots__ = ('__objects',)
 
-    def pk():
-        """Primary key alias
-        """
+    @property
+    def pk(self):
+        """Get the primary key."""
+        if 'id_field' not in self._meta:
+            return None
+        return getattr(self, self._meta['id_field'])
 
-        def fget(self):
-            if 'id_field' not in self._meta:
-                return None
-            return getattr(self, self._meta['id_field'])
-
-        def fset(self, value):
-            return setattr(self, self._meta['id_field'], value)
-
-        return property(fget, fset)
-
-    pk = pk()
+    @pk.setter
+    def pk(self, value):
+        """Set the primary key."""
+        return setattr(self, self._meta['id_field'], value)
 
     @classmethod
     def _get_db(cls):
         """Some Model using other db_alias"""
-        return get_db(cls._meta.get("db_alias", DEFAULT_CONNECTION_NAME))
+        return get_db(cls._meta.get('db_alias', DEFAULT_CONNECTION_NAME))
 
     @classmethod
     def _get_collection(cls):
@@ -209,7 +204,20 @@ class Document(BaseDocument, metaclass=TopLevelDocumentMetaclass):
                 cls.ensure_indexes()
         return cls._collection
 
-    def modify(self, query={}, **update):
+    def to_mongo(self, *args, **kwargs):
+        data = super(Document, self).to_mongo(*args, **kwargs)
+
+        # If '_id' is None, try and set it from self._data. If that
+        # doesn't exist either, remote '_id' from the SON completely.
+        if data['_id'] is None:
+            if self._data.get('id') is None:
+                del data['_id']
+            else:
+                data['_id'] = self._data['id']
+
+        return data
+
+    def modify(self, query=None, **update):
         """Perform an atomic update of the document in the database and reload
         the document object using updated version.
 
@@ -223,17 +231,19 @@ class Document(BaseDocument, metaclass=TopLevelDocumentMetaclass):
             database matches the query
         :param update: Django-style update keyword arguments
         """
+        if query is None:
+            query = {}
 
         if self.pk is None:
-            raise InvalidDocumentError("The document does not have a primary key.")
+            raise InvalidDocumentError('The document does not have a primary key.')
 
-        id_field = self._meta["id_field"]
+        id_field = self._meta['id_field']
         query = query.copy() if isinstance(query, dict) else query.to_query(self)
 
         if id_field not in query:
             query[id_field] = self.pk
         elif query[id_field] != self.pk:
-            raise InvalidQueryError("Invalid document modify query: it must modify only this document.")
+            raise InvalidQueryError('Invalid document modify query: it must modify only this document.')
 
         updated = self._qs(**query).modify(new=True, **update)
         if updated is None:
@@ -249,7 +259,7 @@ class Document(BaseDocument, metaclass=TopLevelDocumentMetaclass):
 
     def save(self, force_insert=False, validate=True, clean=True,
              write_concern=None, cascade=None, cascade_kwargs=None,
-             _refs=None, save_condition=None, **kwargs):
+             _refs=None, save_condition=None, signal_kwargs=None, **kwargs):
         """Save the :class:`~mongoengine.Document` to the database. If the
         document already exists, it will be updated, otherwise it will be
         created.
@@ -275,6 +285,8 @@ class Document(BaseDocument, metaclass=TopLevelDocumentMetaclass):
         :param save_condition: only perform save if matching record in db
             satisfies condition(s) (e.g. version number).
             Raises :class:`OperationError` if the conditions are not satisfied
+        :parm signal_kwargs: (optional) kwargs dictionary to be passed to
+            the signal calls.
 
         .. versionchanged:: 0.5
             In existing documents it only saves changed fields using
@@ -296,21 +308,24 @@ class Document(BaseDocument, metaclass=TopLevelDocumentMetaclass):
             :class:`OperationError` exception raised if save_condition fails.
         .. versionchanged:: 0.10.1
             :class: save_condition failure now raises a `SaveConditionError`
+        .. versionchanged:: 0.10.7
+            Add signal_kwargs argument
         """
-        signals.pre_save.send(self.__class__, document=self)
+        signal_kwargs = signal_kwargs or {}
+        signals.pre_save.send(self.__class__, document=self, **signal_kwargs)
 
         if validate:
             self.validate(clean=clean)
 
         if write_concern is None:
-            write_concern = {"w": 1}
+            write_concern = {'w': 1}
 
         doc = self.to_mongo()
 
         created = ('_id' not in doc or self._created or force_insert)
 
         signals.pre_save_post_validation.send(self.__class__, document=self,
-                                              created=created)
+                                              created=created, **signal_kwargs)
 
         try:
             collection = self._get_collection()
@@ -326,8 +341,10 @@ class Document(BaseDocument, metaclass=TopLevelDocumentMetaclass):
                     # Correct behaviour in 2.X and in 3.0.1+ versions
                     if not object_id and pymongo.version_tuple == (3, 0):
                         pk_as_mongo_obj = self._fields.get(self._meta['id_field']).to_mongo(self.pk)
-                        object_id = self._qs.filter(pk=pk_as_mongo_obj).first() and \
-                                    self._qs.filter(pk=pk_as_mongo_obj).first().pk
+                        object_id = (
+                            self._qs.filter(pk=pk_as_mongo_obj).first() and
+                            self._qs.filter(pk=pk_as_mongo_obj).first().pk
+                        )  # TODO doesn't this make 2 queries?
             else:
                 object_id = doc['_id']
                 updates, removals = self._delta()
@@ -338,7 +355,7 @@ class Document(BaseDocument, metaclass=TopLevelDocumentMetaclass):
                 else:
                     select_dict = {}
                 select_dict['_id'] = object_id
-                shard_key = self.__class__._meta.get('shard_key', tuple())
+                shard_key = self._meta.get('shard_key', tuple())
                 for k in shard_key:
                     path = self._lookup_field(k.split('.'))
                     actual_key = [p.db_field for p in path]
@@ -349,7 +366,7 @@ class Document(BaseDocument, metaclass=TopLevelDocumentMetaclass):
 
                 def is_new_object(last_error):
                     if last_error is not None:
-                        updated = last_error.get("updatedExisting")
+                        updated = last_error.get('updatedExisting')
                         if updated is not None:
                             return not updated
                     return created
@@ -357,14 +374,14 @@ class Document(BaseDocument, metaclass=TopLevelDocumentMetaclass):
                 update_query = {}
 
                 if updates:
-                    update_query["$set"] = updates
+                    update_query['$set'] = updates
                 if removals:
-                    update_query["$unset"] = removals
+                    update_query['$unset'] = removals
                 if updates or removals:
                     upsert = save_condition is None
                     last_error = collection.update(select_dict, update_query,
                                                    upsert=upsert, **write_concern)
-                    if not upsert and last_error["n"] == 0:
+                    if not upsert and last_error['n'] == 0:
                         raise SaveConditionError('Race condition preventing'
                                                  ' document update detected')
                     created = is_new_object(last_error)
@@ -375,10 +392,10 @@ class Document(BaseDocument, metaclass=TopLevelDocumentMetaclass):
 
             if cascade:
                 kwargs = {
-                    "force_insert": force_insert,
-                    "validate": validate,
-                    "write_concern": write_concern,
-                    "cascade": cascade
+                    'force_insert': force_insert,
+                    'validate': validate,
+                    'write_concern': write_concern,
+                    'cascade': cascade
                 }
                 if cascade_kwargs:  # Allow granular control over cascades
                     kwargs.update(cascade_kwargs)
@@ -386,28 +403,31 @@ class Document(BaseDocument, metaclass=TopLevelDocumentMetaclass):
                 self.cascade_save(**kwargs)
         except pymongo.errors.DuplicateKeyError as err:
             message = 'Tried to save duplicate unique keys (%s)'
-            raise NotUniqueError(message % str(err))
+            raise NotUniqueError(message % six.text_type(err))
         except pymongo.errors.OperationFailure as err:
             message = 'Could not save document (%s)'
-            if re.match('^E1100[01] duplicate key', str(err)):
+            if re.match('^E1100[01] duplicate key', six.text_type(err)):
                 # E11000 - duplicate key error index
                 # E11001 - duplicate key on update
                 message = 'Tried to save duplicate unique keys (%s)'
-                raise NotUniqueError(message % str(err))
-            raise OperationError(message % str(err))
+                raise NotUniqueError(message % six.text_type(err))
+            raise OperationError(message % six.text_type(err))
+
         id_field = self._meta['id_field']
         if created or id_field not in self._meta.get('shard_key', []):
             self[id_field] = self._fields[id_field].to_python(object_id)
 
-        signals.post_save.send(self.__class__, document=self, created=created)
+        signals.post_save.send(self.__class__, document=self,
+                               created=created, **signal_kwargs)
         self._clear_changed_fields()
         self._created = False
         return self
 
-    def cascade_save(self, *args, **kwargs):
-        """Recursively saves any references /
-           generic references on the document"""
-        _refs = kwargs.get('_refs', []) or []
+    def cascade_save(self, **kwargs):
+        """Recursively save any references and generic references on the
+        document.
+        """
+        _refs = kwargs.get('_refs') or []
 
         ReferenceField = _import_class('ReferenceField')
         GenericReferenceField = _import_class('GenericReferenceField')
@@ -433,16 +453,17 @@ class Document(BaseDocument, metaclass=TopLevelDocumentMetaclass):
 
     @property
     def _qs(self):
-        """
-        Returns the queryset to use for updating / reloading / deletions
-        """
+        """Return the queryset to use for updating / reloading / deletions."""
         if not hasattr(self, '__objects'):
             self.__objects = QuerySet(self, self._get_collection())
         return self.__objects
 
     @property
     def _object_key(self):
-        """Dict to identify object in collection
+        """Get the query dict that can be used to fetch this object from
+        the database. Most of the time it's a simple PK lookup, but in
+        case of a sharded collection with a compound shard key, it can
+        contain a more complex query.
         """
         select_dict = {'pk': self.pk}
         shard_key = self.__class__._meta.get('shard_key', tuple())
@@ -462,11 +483,11 @@ class Document(BaseDocument, metaclass=TopLevelDocumentMetaclass):
         Raises :class:`OperationError` if called on an object that has not yet
         been saved.
         """
-        if not self.pk:
+        if self.pk is None:
             if kwargs.get('upsert', False):
                 query = self.to_mongo()
-                if "_cls" in query:
-                    del query["_cls"]
+                if '_cls' in query:
+                    del query['_cls']
                 return self._qs.filter(**query).update_one(**kwargs)
             else:
                 raise OperationError(
@@ -475,23 +496,29 @@ class Document(BaseDocument, metaclass=TopLevelDocumentMetaclass):
         # Need to add shard key to query, or you get an error
         return self._qs.filter(**self._object_key).update_one(**kwargs)
 
-    def delete(self, **write_concern):
+    def delete(self, signal_kwargs=None, **write_concern):
         """Delete the :class:`~mongoengine.Document` from the database. This
         will only take effect if the document has been previously saved.
 
+        :parm signal_kwargs: (optional) kwargs dictionary to be passed to
+            the signal calls.
         :param write_concern: Extra keyword arguments are passed down which
             will be used as options for the resultant
             ``getLastError`` command.  For example,
             ``save(..., write_concern={w: 2, fsync: True}, ...)`` will
             wait until at least two servers have recorded the write and
             will force an fsync on the primary server.
-        """
-        signals.pre_delete.send(self.__class__, document=self)
 
-        # Delete FileFields separately 
+        .. versionchanged:: 0.10.7
+            Add signal_kwargs argument
+        """
+        signal_kwargs = signal_kwargs or {}
+        signals.pre_delete.send(self.__class__, document=self, **signal_kwargs)
+
+        # Delete FileFields separately
         FileField = _import_class('FileField')
         for name, field in self._fields.items():
-            if isinstance(field, FileField): 
+            if isinstance(field, FileField):
                 getattr(self, name).delete()
 
         try:
@@ -500,7 +527,7 @@ class Document(BaseDocument, metaclass=TopLevelDocumentMetaclass):
         except pymongo.errors.OperationFailure as err:
             message = 'Could not delete document (%s)' % err.message
             raise OperationError(message)
-        signals.post_delete.send(self.__class__, document=self)
+        signals.post_delete.send(self.__class__, document=self, **signal_kwargs)
 
     def switch_db(self, db_alias, keep_created=True):
         """
@@ -585,11 +612,12 @@ class Document(BaseDocument, metaclass=TopLevelDocumentMetaclass):
         if fields and isinstance(fields[0], int):
             max_depth = fields[0]
             fields = fields[1:]
-        elif "max_depth" in kwargs:
-            max_depth = kwargs["max_depth"]
+        elif 'max_depth' in kwargs:
+            max_depth = kwargs['max_depth']
+
+        if self.pk is None:
+            raise self.DoesNotExist('Document does not exist')
 
-        if not self.pk:
-            raise self.DoesNotExist("Document does not exist")
         obj = self._qs.read_preference(ReadPreference.PRIMARY).filter(
             **self._object_key).only(*fields).limit(
             1).select_related(max_depth=max_depth)
@@ -597,7 +625,7 @@ class Document(BaseDocument, metaclass=TopLevelDocumentMetaclass):
         if obj:
             obj = obj[0]
         else:
-            raise self.DoesNotExist("Document does not exist")
+            raise self.DoesNotExist('Document does not exist')
 
         for field in obj._data:
             if not fields or field in fields:
@@ -639,8 +667,8 @@ class Document(BaseDocument, metaclass=TopLevelDocumentMetaclass):
     def to_dbref(self):
         """Returns an instance of :class:`~bson.dbref.DBRef` useful in
         `__raw__` queries."""
-        if not self.pk:
-            msg = "Only saved documents can have a valid dbref"
+        if self.pk is None:
+            msg = 'Only saved documents can have a valid dbref'
             raise OperationError(msg)
         return DBRef(self.__class__._get_collection_name(), self.pk)
 
@@ -666,10 +694,20 @@ class Document(BaseDocument, metaclass=TopLevelDocumentMetaclass):
     def drop_collection(cls):
         """Drops the entire collection associated with this
         :class:`~mongoengine.Document` type from the database.
+
+        Raises :class:`OperationError` if the document has no collection set
+        (i.g. if it is `abstract`)
+
+        .. versionchanged:: 0.10.7
+            :class:`OperationError` exception raised if no collection available
         """
+        col_name = cls._get_collection_name()
+        if not col_name:
+            raise OperationError('Document %s has no collection defined '
+                                 '(is it abstract ?)' % cls)
         cls._collection = None
         db = cls._get_db()
-        db.drop_collection(cls._get_collection_name())
+        db.drop_collection(col_name)
 
     @classmethod
     def create_index(cls, keys, background=False, **kwargs):
@@ -685,7 +723,7 @@ class Document(BaseDocument, metaclass=TopLevelDocumentMetaclass):
         fields = index_spec.pop('fields')
         drop_dups = kwargs.get('drop_dups', False)
         if IS_PYMONGO_3 and drop_dups:
-            msg = "drop_dups is deprecated and is removed when using PyMongo 3+."
+            msg = 'drop_dups is deprecated and is removed when using PyMongo 3+.'
             warnings.warn(msg, DeprecationWarning)
         elif not IS_PYMONGO_3:
             index_spec['drop_dups'] = drop_dups
@@ -711,7 +749,7 @@ class Document(BaseDocument, metaclass=TopLevelDocumentMetaclass):
             will be removed if PyMongo3+ is used
         """
         if IS_PYMONGO_3 and drop_dups:
-            msg = "drop_dups is deprecated and is removed when using PyMongo 3+."
+            msg = 'drop_dups is deprecated and is removed when using PyMongo 3+.'
             warnings.warn(msg, DeprecationWarning)
         elif not IS_PYMONGO_3:
             kwargs.update({'drop_dups': drop_dups})
@@ -731,7 +769,7 @@ class Document(BaseDocument, metaclass=TopLevelDocumentMetaclass):
         index_opts = cls._meta.get('index_opts') or {}
         index_cls = cls._meta.get('index_cls', True)
         if IS_PYMONGO_3 and drop_dups:
-            msg = "drop_dups is deprecated and is removed when using PyMongo 3+."
+            msg = 'drop_dups is deprecated and is removed when using PyMongo 3+.'
             warnings.warn(msg, DeprecationWarning)
 
         collection = cls._get_collection()
@@ -769,8 +807,7 @@ class Document(BaseDocument, metaclass=TopLevelDocumentMetaclass):
 
         # If _cls is being used (for polymorphism), it needs an index,
         # only if another index doesn't begin with _cls
-        if (index_cls and not cls_indexed and
-                cls._meta.get('allow_inheritance', ALLOW_INHERITANCE) is True):
+        if index_cls and not cls_indexed and cls._meta.get('allow_inheritance'):
 
             # we shouldn't pass 'cls' to the collection.ensureIndex options
             # because of https://jira.mongodb.org/browse/SERVER-769
@@ -840,16 +877,15 @@ class Document(BaseDocument, metaclass=TopLevelDocumentMetaclass):
         # finish up by appending { '_id': 1 } and { '_cls': 1 }, if needed
         if [('_id', 1)] not in indexes:
             indexes.append([('_id', 1)])
-        if (cls._meta.get('index_cls', True) and
-                cls._meta.get('allow_inheritance', ALLOW_INHERITANCE) is True):
+        if cls._meta.get('index_cls', True) and cls._meta.get('allow_inheritance'):
             indexes.append([('_cls', 1)])
 
         return indexes
 
     @classmethod
     def compare_indexes(cls):
-        """ Compares the indexes defined in MongoEngine with the ones existing
-        in the database. Returns any missing/extra indexes.
+        """ Compares the indexes defined in MongoEngine with the ones
+        existing in the database. Returns any missing/extra indexes.
         """
 
         required = cls.list_indexes()
@@ -892,8 +928,9 @@ class DynamicDocument(Document, metaclass=TopLevelDocumentMetaclass):
     _dynamic = True
 
     def __delattr__(self, *args, **kwargs):
-        """Deletes the attribute by setting to None and allowing _delta to unset
-        it"""
+        """Delete the attribute by setting to None and allowing _delta
+        to unset it.
+        """
         field_name = args[0]
         if field_name in self._dynamic_fields:
             setattr(self, field_name, None)
@@ -914,12 +951,13 @@ class DynamicEmbeddedDocument(EmbeddedDocument, metaclass=DocumentMetaclass):
     _dynamic = True
 
     def __delattr__(self, *args, **kwargs):
-        """Deletes the attribute by setting to None and allowing _delta to unset
-        it"""
+        """Delete the attribute by setting to None and allowing _delta
+        to unset it.
+        """
         field_name = args[0]
         if field_name in self._fields:
             default = self._fields[field_name].default
-            if isinstance(default, collections.Callable):
+            if callable(default):
                 default = default()
             setattr(self, field_name, default)
         else:
@@ -956,11 +994,11 @@ class MapReduceDocument(object):
         if not isinstance(self.key, id_field_type):
             try:
                 self.key = id_field_type(self.key)
-            except:
-                raise Exception("Could not cast key as %s" %
+            except Exception:
+                raise Exception('Could not cast key as %s' %
                                 id_field_type.__name__)
 
-        if not hasattr(self, "_key_object"):
+        if not hasattr(self, '_key_object'):
             self._key_object = self._document.objects.with_id(self.key)
             return self._key_object
         return self._key_object
